#include "render.ph"

RenderCtx* render_init(bool enable_validation_layers)
{
	RenderCtx* ctx = (RenderCtx*) calloc(1, sizeof(RenderCtx));
	CK_PTR(ctx);

	ctx->enable_validation_layers = enable_validation_layers;

	CK_BOOL(render_init_layers(ctx));
	CK_BOOL(render_init_extensions(ctx));
	CK_BOOL(render_init_instance(ctx));

	LOG_I("Vulkan Instance created");

	CK_BOOL(render_init_procs(ctx));

	if (enable_validation_layers)
	{
		CK_BOOL(render_init_debug_messenger(ctx));
	}

	CK_BOOL(render_enumerate_physical_devices(ctx));
	CK_BOOL(render_enumerate_queue_family_properties(ctx, ctx->physical_devices[ctx->active_physical_device]));
	CK_BOOL(render_find_queue_families(ctx));
	CK_BOOL(render_init_device_extensions(ctx));
	CK_BOOL(render_init_logical_device(ctx));

	LOG_I("Vulkan Logical Device created");

	return ctx;
error:
	return NULL;
}

bool render_init_layers(RenderCtx* ctx)
{
	CK_PTR(ctx);

	CK_BOOL(render_enumerate_layers(ctx));

	if (ctx->enable_validation_layers)
	{
		render_push_layer(ctx, "VK_LAYER_KHRONOS_validation");
	}

	return true;
error:
	return false;
}

bool render_init_extensions(RenderCtx* ctx)
{
	CK_PTR(ctx);

	CK_BOOL(render_enumerate_extensions(ctx, NULL)); // Base layer

	if (ctx->layers && ctx->enable_validation_layers)
	{
		CK_BOOL(render_enumerate_extensions(ctx, ctx->layers[0])); // Validation layers

		render_push_extension(ctx, VK_EXT_DEBUG_UTILS_EXTENSION_NAME);
	}

	return true;
error:
	return false;
}

bool render_init_device_extensions(RenderCtx* ctx)
{
	CK_PTR(ctx);

	CK_BOOL(render_enumerate_device_extensions(ctx,ctx->physical_devices[ctx->active_physical_device]));

	render_push_device_extension(ctx, VK_KHR_SWAPCHAIN_EXTENSION_NAME);
	render_push_device_extension(ctx, VK_KHR_EXTERNAL_MEMORY_FD_EXTENSION_NAME);
	render_push_device_extension(ctx, VK_EXT_EXTERNAL_MEMORY_DMA_BUF_EXTENSION_NAME);

	return true;
error:
	return false;
}

bool render_init_instance(RenderCtx* ctx)
{
	CK_PTR(ctx);

	VkResult result;
	VkApplicationInfo app_info = {0};
	VkInstanceCreateInfo create_info = {0};
	VkDebugUtilsMessengerCreateInfoEXT debug_create_info = {0};

	debug_create_info = render_get_debug_create_info();

	app_info.sType = VK_STRUCTURE_TYPE_APPLICATION_INFO;
	app_info.pApplicationName = APPLICATION_NAME;
	app_info.applicationVersion = VK_MAKE_VERSION(VERSION_MAJOR, VERSION_MINOR, VERSION_PATCH);
	app_info.pEngineName = ENGINE_NAME;
	app_info.engineVersion = VK_MAKE_VERSION(VERSION_MAJOR, VERSION_MINOR, VERSION_PATCH);
	app_info.apiVersion = VK_API_VERSION_1_2;
	// app_info.apiVersion = VK_HEADER_VERSION_COMPLETE;

	create_info.sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO;
	create_info.pApplicationInfo = &app_info;
	create_info.enabledLayerCount = ctx->layers != NULL ? ctx->layers_count : 0;
	create_info.ppEnabledLayerNames = ctx->layers != NULL ? (const char* const*) ctx->layers : NULL;
	create_info.enabledExtensionCount = ctx->extensions != NULL ? ctx->extensions_count : 0;
	create_info.ppEnabledExtensionNames = ctx->extensions != NULL ? (const char* const*) ctx->extensions : NULL;

	if (ctx->enable_validation_layers)
	{
		create_info.pNext = &debug_create_info; // Enables validation layers for vkCreateInstance
	}

	result = vkCreateInstance(&create_info, ctx->allocators, &ctx->instance);
	CK_VK(result);

	return true;
error:
	return false;
}

#define QUEUE_COUNT 1
bool render_init_logical_device(RenderCtx* ctx)
{
	VkResult result;
	float queue_priorities = 1.0f;
	VkDeviceQueueCreateInfo* queue_create_infos = NULL;
	VkPhysicalDeviceFeatures device_features = {0};
	VkDeviceCreateInfo device_create_info = {0};

	CK_PTR(ctx);

	// Queues
	queue_create_infos = (VkDeviceQueueCreateInfo*) calloc(QUEUE_COUNT, sizeof(VkDeviceQueueCreateInfo));

	queue_create_infos[0].sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO;
	queue_create_infos[0].queueFamilyIndex = ctx->queue_indices.graphics;
	queue_create_infos[0].queueCount = 1;
	queue_create_infos[0].pQueuePriorities = &queue_priorities;

	// queue_create_infos[1].sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO;
	// queue_create_infos[1].queueFamilyIndex = ctx->queue_indices.present;
	// queue_create_infos[1].queueCount = 1;
	// queue_create_infos[1].pQueuePriorities = &queue_priorities;

	// Device
	device_create_info.sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO;
	device_create_info.pEnabledFeatures = &device_features;
	device_create_info.enabledLayerCount = ctx->layers != NULL ? ctx->layers_count : 0;
	device_create_info.ppEnabledLayerNames = ctx->layers != NULL ? (const char* const*) ctx->layers : NULL;
	device_create_info.enabledExtensionCount = ctx->device_extensions != NULL ? ctx->device_extensions_count : 0;
	device_create_info.ppEnabledExtensionNames = ctx->device_extensions != NULL ? (const char* const*) ctx->device_extensions : NULL;
	device_create_info.queueCreateInfoCount = QUEUE_COUNT;
	device_create_info.pQueueCreateInfos = queue_create_infos;

	result = vkCreateDevice(ctx->physical_devices[ctx->active_physical_device], &device_create_info, NULL, &ctx->device);
	CK_VK(result);

	// vkGetDeviceQueue(ctx->device, ctx->queue_indices.graphics, 0, &ctx->graphics_queue);
	// vkGetDeviceQueue(ctx->device, ctx->queue_indices.present, 0, &ctx->present_queue);

	free(queue_create_infos);
	return true;
error:
	free(queue_create_infos);
	return false;
}
